import urllib.request
import urllib.error
import typing
from email.message import Message

class Response(typing.NamedTuple):
    content: bytes
    head: Message
    status_code: int
    @property
    def headers(self):
        res={}
        h=self.head.items()
        for item in h:
            res[item[0]]=item[1]
        return res


def request(
    url: str
) -> Response:
    method = 'GET'

    headers = {"Accept": "*"}

    httprequest = urllib.request.Request(
        url,  headers=headers, method=method
    )

    try:
        with urllib.request.urlopen(httprequest) as httpresponse:
            response = Response(
                head=httpresponse.headers,
                status_code=httpresponse.status,
                content=httpresponse.read()
            )
    except urllib.error.HTTPError as e:
        response = Response(
            content=bytes(e.reason),
            head=e.headers,
            status_code=e.code
        )

    return response



import os
import zipfile
from datetime import datetime

exclude_all = [
    '__pycache__',
    'node_modules',
    '.idea',
    '.git',
    '.github'
]
exclude = [
    'storage/cache',
    'storage/logs'
]


def zipdir(path, ziph):
    for dirname, subdirs, files in os.walk(path):
        has=False
        for dir in exclude_all:
            if dirname.find(dir)!=-1:
                has=True
                break
        for dir in exclude:
            if dirname.startswith(dir):
                has = True
                break
        if has:
            continue
        ziph.write(dirname)
        for file in files:
            ziph.write(os.path.join(dirname, file))

def folders():
    res=[]
    for entry in os.scandir('./'):
        if entry.is_dir():
            res.append(entry.name)

    return res

def run():
    path = input('backups folder path:<D:/projects/backups>').lower()
    if path =='':
        path='D:/projects/backups'

    ff=folders()
    date = str(datetime.utcnow().date())
    app_name = input('backup name:')
    print(f'start backing up {app_name}...')
    if not os.path.exists(f'{path}/{app_name}'):
        os.makedirs(f'{path}/{app_name}')
    zipf = zipfile.ZipFile(f'{path}/{app_name}/{app_name}_{date}.zip', 'w', zipfile.ZIP_DEFLATED)
    for folder in ff:
        if folder not in exclude_all:
            zipdir(f'{folder}/', zipf)

    zipf.write('index.py')
    zipf.close()

    print('end')




import typing
import re

try:
    from bson import ObjectId
except:
    def ObjectId(s):
        return s


def full_path_builder(app_prefix: str, path: str):
    app_prefix = app_prefix.strip(' /')
    path = path.strip(' /')
    full = ''
    if app_prefix != '':
        full += '/' + app_prefix
    full = full.strip(' /')
    full = full.strip(' /')
    if path != '':
        full += '/' + path
    return '/' + full.strip(' /')


class BaseRoute:
    _routes = {}
    _route_store = {}

    def __init__(self, scope, prefix: str = '', middlewares=None, sub_domain=None) -> None:
        if middlewares is None:
            middlewares = []

        self._scope = scope
        self._app_prefix = prefix
        self._middlewares = middlewares
        self._sub_domain = sub_domain or '_'

        if self._sub_domain not in self._routes:
            self._routes[self._sub_domain] = {}

        for m in ['POST', 'GET', 'PUT', 'OPTIONS', 'DELETE', 'WS']:
            if m not in self._routes[self._sub_domain]:
                self._routes[self._sub_domain][m] = []

    def _add(self, path: str, controller: typing.Callable, func: typing.Union[str, typing.Callable], method: str,
             middlewares=None, cache=None):
        if middlewares is None:
            middlewares = []
        full_path = full_path_builder(self._app_prefix, path)
        middlwrs = self._middlewares.copy()

        r = {
            'path': full_path,
            'controller': controller,
            'func': func,
            'middlewares': middlwrs + middlewares,
            'regex': build_path(full_path)
        }
        if cache:
            r['cache'] = cache
        self._routes[self._sub_domain][method].append(r)

    @property
    def all(self):
        return self._routes

    @property
    def sub_domains(self):
        if self._route_store.get('_subdomains', False) is False:
            self._route_store['_subdomains'] = list(self._routes.keys())

        return self._route_store['_subdomains']

    def __request_subdomain(self, request):
        if len(self.sub_domains) == 1:
            return self.sub_domains[0]

        return request.sub_domain or '_'

    def response(self, request):
        method = request.method
        if method not in ['POST', 'GET', 'PUT', 'OPTIONS', 'DELETE', 'WS']:
            return False

        if self._scope['path'] != '/':
            self._scope['path'] = self._scope['path'].rstrip(' /')

        path = self._scope['path']

        sub_domain = self.__request_subdomain(request)

        if sub_domain not in self._routes:
            return False
        for route in self._routes[sub_domain][method]:
            regex, params = route['regex']
            find = regex.search(path)
            if find:
                args = {}
                for key, value in params.items():
                    args[key] = value.convert(find.group(key))
                return build(route, args)

        return False


class Router(BaseRoute):
    def __init__(self, scope=None, prefix: str = '', middlewares: typing.List[typing.Callable] = None,
                 sub_domain: str = None) -> None:
        if middlewares is None:
            middlewares = []
        super().__init__(scope, prefix, middlewares, sub_domain)

    def get(self, path, controller: typing.Callable = None, func: typing.Union[str, typing.Callable] = None,
            middlewares: typing.List[typing.Callable] = None, cache=None):
        self._add(path, controller, func, 'GET', middlewares, cache)
        return self

    def post(self, path, controller: typing.Callable = None, func: typing.Union[str, typing.Callable] = None,
             middlewares: typing.List[typing.Callable] = None):
        self._add(path, controller, func, 'POST', middlewares)
        return self

    def put(self, path, controller: typing.Callable = None, func: typing.Union[str, typing.Callable] = None,
            middlewares: typing.List[typing.Callable] = None):
        self._add(path, controller, func, 'PUT', middlewares)
        return self

    def option(self, path, controller: typing.Callable = None, func: typing.Union[str, typing.Callable] = None,
               middlewares: typing.List[typing.Callable] = None):
        self._add(path, controller, func, 'OPTIONS', middlewares)
        return self

    def delete(self, path, controller: typing.Callable = None, func: typing.Union[str, typing.Callable] = None,
               middlewares: typing.List[typing.Callable] = None):
        self._add(path, controller, func, 'DELETE', middlewares)
        return self

    def ws(self, path, controller: typing.Callable = None, func: typing.Union[str, typing.Callable] = None,
           middlewares: typing.List[typing.Callable] = None):
        self._add(path, controller, func, 'WS', middlewares)
        return self

    def crud(self, path, controller: typing.Callable = None, middlewares: typing.List[typing.Callable] = None):
        self._add(path, controller, 'index', 'GET', middlewares)
        self._add(path, controller, 'store', 'POST', middlewares)
        self._add(path + '/{id:oid}', controller, 'update', 'PUT', middlewares)
        self._add(path + '/{id:oid}', controller, 'delete', 'DELETE', middlewares)
        return self

    def mcud(self, path, controller: typing.Callable = None, middlewares: typing.List[typing.Callable] = None):
        self._add(path, controller, 'm_store', 'POST', middlewares)
        self._add(path, controller, 'm_update', 'PUT', middlewares)
        self._add(path, controller, 'm_delete', 'DELETE', middlewares)
        return self


def build(route, args):
    res = {}
    res['path'] = route['path']
    res['args'] = args
    res['controller'] = route['controller']
    res['func'] = route['func']
    res['middlewares'] = route['middlewares']
    res['cache'] = route.get('cache', None)
    return res


PARAM_REGEX = re.compile("{([a-zA-Z_][a-zA-Z0-9_]*)(:[^/{]*)?}")

class StringConvertor:
    regex = "[^/]+"

    def convert(self, value: str) -> str:
        return value


class PathConvertor:
    regex = ".*"

    def convert(self, value: str) -> str:
        return str(value)


class IntegerConvertor:
    regex = "[0-9]+"

    def convert(self, value: str) -> int:
        return int(value)


class FloatConvertor:
    regex = "[0-9]+(.[0-9]+)?"

    def convert(self, value: str) -> float:
        return float(value)


class OIdConvertor:
    regex = "[a-f\d]{24}"

    def convert(self, value: str) -> ObjectId:
        return ObjectId(value)


class BoolConvertor:
    regex = "(true)|(false)"

    def convert(self, value: str) -> bool:
        return bool(value)


CONVERTOR_TYPES = {
    "str": StringConvertor(),
    "int": IntegerConvertor(),
    "float": FloatConvertor(),
    "path": PathConvertor(),
    "oid": OIdConvertor(),
    "bool": BoolConvertor()
}


def build_path(
        path: str,
):
    path_regex = "^"
    idx = 0
    param_convertors = {}
    for match in PARAM_REGEX.finditer(path):
        param_name, convertor_type = match.groups("str")
        convertor_type = convertor_type.lstrip(":")
        path_regex += re.escape(path[idx: match.start()])
        if convertor_type in CONVERTOR_TYPES:
            convertor = CONVERTOR_TYPES[convertor_type]
            path_regex += f"(?P<{param_name}>{convertor.regex})"
            param_convertors[param_name] = convertor
        else:
            path_regex += f"(?P<{param_name}>{convertor_type})"
            param_convertors[param_name] = CONVERTOR_TYPES['str']

        idx = match.end()

    path_regex += re.escape(path[idx:]) + "$"

    return re.compile(path_regex), param_convertors




from base64 import urlsafe_b64encode, urlsafe_b64decode
import sys

from renus.util.helper import get_random_string, remove_pad64, add_pad64

try:
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
    from cryptography.fernet import Fernet
except ImportError:
    Fernet = None

def pad(s,n=16):
    block_size = n
    remainder = len(s) % block_size
    padding_needed = block_size - remainder
    return s + padding_needed * ' '


def __generate_key(password):
    salt = b'dfkTGG536fdfkTGG536fkdfijdkdfijd'
    kdf = PBKDF2HMAC(algorithm = hashes.SHA256(),length = 32,salt = salt,iterations = 100000)
    return urlsafe_b64encode(kdf.derive(password.encode()))


def encrypt(text:str,password:str)->str:
    assert Fernet is not None, "'cryptography' must be installed to use encrypt"
    key = __generate_key(password)
    f = Fernet(key)
    d= f.encrypt(text.encode())
    return d.decode()



def decrypt(enc_dict:str, password:str)->str:
    assert Fernet is not None, "'cryptography' must be installed to use decrypt"
    key = __generate_key(password)
    f = Fernet(key)
    d = f.decrypt(enc_dict.encode())
    return d.decode()

def shift(text:str,password:str):
    rnd = get_random_string(6)
    password = rnd + password
    password = urlsafe_b64encode(password.encode()).decode()

    text=urlsafe_b64encode(text.encode()).decode()
    text = pad(text, len(password))
    lp=[]
    ls=[]
    for p in password:
        lp.append(ord(p))
    ii = 0
    def nx(ii):
        if len(lp)<ii+2:
            return 0
        return ii+1

    for t in text:
        ls.append(chr(ord(t) + lp[ii]))
        ii = nx(ii)

    text=rnd+''.join(ls)
    text=urlsafe_b64encode(text.encode()).decode()
    return remove_pad64(text)

def unshift(text: str, password: str):
    try:
        text=add_pad64(text)
        text = urlsafe_b64decode(text).decode()
        rnd = text[:6]
        text = text[6:]
        password = rnd + password
        password = urlsafe_b64encode(password.encode()).decode()
        lp = []
        ls = []
        for p in password:
            lp.append(ord(p))
        ii = 0

        def nx(ii):
            if len(lp) < ii + 2:
                return 0
            return ii + 1

        for t in text:
            ls.append(chr(ord(t) - lp[ii]))
            ii = nx(ii)

        if len(ls)<len(lp):
            raise RuntimeError('Invalid Token Format')

        text=''.join(ls).rstrip()
        return urlsafe_b64decode(text).decode()
    except:
        sys.exit('Invalid Token Format')


class Status:

    """
    HTTP codes
    See RFC 2616 - https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
    And RFC 6585 - https://tools.ietf.org/html/rfc6585
    And RFC 4918 - https://tools.ietf.org/html/rfc4918
    And RFC 8470 - https://tools.ietf.org/html/rfc8470
    """
    HTTP_100_CONTINUE = 100
    HTTP_101_SWITCHING_PROTOCOLS = 101
    HTTP_200_OK = 200
    HTTP_201_CREATED = 201
    HTTP_202_ACCEPTED = 202
    HTTP_203_NON_AUTHORITATIVE_INFORMATION = 203
    HTTP_204_NO_CONTENT = 204
    HTTP_205_RESET_CONTENT = 205
    HTTP_206_PARTIAL_CONTENT = 206
    HTTP_207_MULTI_STATUS = 207
    HTTP_300_MULTIPLE_CHOICES = 300
    HTTP_301_MOVED_PERMANENTLY = 301
    HTTP_302_FOUND = 302
    HTTP_303_SEE_OTHER = 303
    HTTP_304_NOT_MODIFIED = 304
    HTTP_305_USE_PROXY = 305
    HTTP_306_RESERVED = 306
    HTTP_307_TEMPORARY_REDIRECT = 307
    HTTP_400_BAD_REQUEST = 400
    HTTP_401_UNAUTHORIZED = 401
    HTTP_402_PAYMENT_REQUIRED = 402
    HTTP_403_FORBIDDEN = 403
    HTTP_404_NOT_FOUND = 404
    HTTP_405_METHOD_NOT_ALLOWED = 405
    HTTP_406_NOT_ACCEPTABLE = 406
    HTTP_407_PROXY_AUTHENTICATION_REQUIRED = 407
    HTTP_408_REQUEST_TIMEOUT = 408
    HTTP_409_CONFLICT = 409
    HTTP_410_GONE = 410
    HTTP_411_LENGTH_REQUIRED = 411
    HTTP_412_PRECONDITION_FAILED = 412
    HTTP_413_REQUEST_ENTITY_TOO_LARGE = 413
    HTTP_414_REQUEST_URI_TOO_LONG = 414
    HTTP_415_UNSUPPORTED_MEDIA_TYPE = 415
    HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE = 416
    HTTP_417_EXPECTATION_FAILED = 417
    HTTP_422_UNPROCESSABLE_ENTITY = 422
    HTTP_423_LOCKED = 423
    HTTP_424_FAILED_DEPENDENCY = 424
    HTTP_425_TOO_EARLY = 425
    HTTP_428_PRECONDITION_REQUIRED = 428
    HTTP_429_TOO_MANY_REQUESTS = 429
    HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE = 431
    HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS = 451
    HTTP_500_INTERNAL_SERVER_ERROR = 500
    HTTP_501_NOT_IMPLEMENTED = 501
    HTTP_502_BAD_GATEWAY = 502
    HTTP_503_SERVICE_UNAVAILABLE = 503
    HTTP_504_GATEWAY_TIMEOUT = 504
    HTTP_505_HTTP_VERSION_NOT_SUPPORTED = 505
    HTTP_507_INSUFFICIENT_STORAGE = 507
    HTTP_511_NETWORK_AUTHENTICATION_REQUIRED = 511

    """
    WebSocket codes
    https://www.iana.org/assignments/websocket/websocket.xml#close-code-number
    https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent
    """


    WS_1000_NORMAL_CLOSURE = 1000
    WS_1001_GOING_AWAY = 1001
    WS_1002_PROTOCOL_ERROR = 1002
    WS_1003_UNSUPPORTED_DATA = 1003
    WS_1004_NO_STATUS_RCVD = 1004
    WS_1005_ABNORMAL_CLOSURE = 1005
    WS_1007_INVALID_FRAME_PAYLOAD_DATA = 1007
    WS_1008_POLICY_VIOLATION = 1008
    WS_1009_MESSAGE_TOO_BIG = 1009
    WS_1010_MANDATORY_EXT = 1010
    WS_1011_INTERNAL_ERROR = 1011
    WS_1012_SERVICE_RESTART = 1012
    WS_1013_TRY_AGAIN_LATER = 1013
    WS_1014_BAD_GATEWAY = 1014
    WS_1015_TLS_HANDSHAKE = 1015
